from __future__ import annotations

import base64
import mimetypes
import os
import zipfile
from io import BytesIO
from pathlib import Path

from js import console, document


def download_path(filename: str, path: Path, mime_type: str = None):
    download_bytes(filename, path.read_bytes(), mime_type)


def download_bytes(filename: str, content: bytes, mime_type: str = None):
    if mime_type is None:
        gt = mimetypes.guess_type(filename)
        mime_type = gt[0]
        if mime_type is None:
            mime_type = 'application/octet-stream'
        console.log(f'guess mime type for `{filename}` is `{mime_type}`')
    a = document.createElement('a')
    a.download = filename
    a.href = f'data:{mime_type};base64,{base64.b64encode(content).decode("ascii")}'
    document.body.append(a)
    a.click()


def _zip_path(zip_file, path):
    path = str(path)
    for root, dirs, files in os.walk(path):
        for file in files:
            zip_file.write(os.path.join(root, file),
                           os.path.relpath(os.path.join(root, file),
                                           os.path.join(path, '../remote')))


def zip_in_memory(path):
    stream = BytesIO()
    with zipfile.ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1) as zip_file:
        _zip_path(zip_file, path)

    stream.seek(0)
    return stream.getbuffer().tobytes()



import errno
import socket

_start = (10000 - 1)
_end = 30000


def find_port() -> int:
    global _start
    while _start < _end:
        _start += 1
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(('0.0.0.0', _start))
                print(f'findport() -> {_start}')
                return _start
            except socket.error as e:
                if e.errno != errno.EADDRINUSE:
                    raise

    raise Exception(f'find_port(start={_start},end={_end}) no bindable tcp port found in interval')


if __name__ == '__main__':
    print(find_port())
    print(find_port())



import inspect
from typing import TypeVar, Callable, List

from js import document, console, HTMLElement
from pyodide.ffi import create_proxy

T = TypeVar('T')


class WidgetProperty:
    def __init__(self, constructor: Callable[[], 'Widget']):
        self.name = ''
        self.constructor = constructor


class Widget:
    def __init__(self, html: str):
        self.html = html
        self._widget_expanded = False
        self._container: HTMLElement = None
        self.holder: HolderWidget = None

    def __call__(self, fun: Callable[[], T]) -> T:
        return WidgetProperty(fun)

    @property
    def container(self) -> HTMLElement:
        if self._container is None:
            self._container = document.createElement('div')
        self._container.setAttribute('w-type', self.__class__.__name__)
        if not self._widget_expanded:
            self._widget_expanded = True
            self._container.innerHTML = self.html
            self.bind_self_elements()
            self._bind_events()
            self.after_render()

        return self._container

    def append_to(self, element: HTMLElement):
        element.append(self.container)
        return self

    def after_render(self):
        pass

    def bind_self_elements(self):
        def get_binder(value):
            if value == self:
                return lambda element: element
            if isinstance(value, WidgetProperty):
                def wc_binder(element):
                    instance: Widget = value.constructor()
                    instance._container = element
                    c = instance.container  # force expand
                    return instance

                return wc_binder
            return None

        for key, value in vars(self).items():
            binder = get_binder(value)
            if binder is None:
                continue

            element = self.container.querySelector('#' + key)
            if element is not None:
                # raise Exception(f'Element not found, name:[{key}] html: [{self.html}]')
                instance = binder(element)
                self.__dict__[key] = instance

    def _bind_events(self):

        members = inspect.getmembers(self)

        for name, method in members:
            parts = name.split('__')
            if len(parts) != 2:
                continue

            element_name = parts[0]
            event_name = parts[1]
            if element_name == '' or event_name == '':
                continue

            element = self.container.querySelector('#' + element_name)
            if element is None:
                console.warn(f'Event bind failed, element `{element_name}` was not found for method `{name}`')
                continue

            m = getattr(self, name)
            element.addEventListener(event_name, create_proxy(m))

    def close(self):
        h = self.holder
        if h is not None:
            h.close(self)

    def uninitialized(self) -> List[WidgetProperty]:
        self_vars = vars(self)
        result = []
        for name, value in self_vars.items():
            if isinstance(value, WidgetProperty):
                value: WidgetProperty
                result.append(value)
                if value.name == '':
                    value.name = name
        return result

    def uninitialized_append_to(self, element: HTMLElement):
        for prop in self.uninitialized():
            instance = prop.constructor()
            setattr(self, prop.name, instance)
            instance.append_to(element)


class HolderWidget(Widget):
    def __init__(self, html: str = ''):
        super().__init__(html)
        self.stack = []
        self.on_show = lambda w: ...

    def show(self, widget: Widget):
        widget.holder = self
        c = self.container
        while c.hasChildNodes():
            c.removeChild(c.firstChild)

        self._remove(widget)
        self.stack.append(widget)
        widget.append_to(self.container)
        self.on_show(widget)

    def close(self, widget: Widget):
        self._remove(widget)
        self.show(self.stack[-1])

    def _remove(self, widget):
        s = self.stack
        if widget in s:
            s.remove(widget)



from typing import Iterable, TypeVar, Generic, Callable, Iterator

_T = TypeVar('_T')


class repeatable_chain(Generic[_T]):
    def __init__(self, *iterables: Iterable[_T]):
        self._iterables = iterables

    def __iter__(self) -> Iterator[_T]:
        def it() -> Iterator[_T]:
            for iterable in self._iterables:
                yield from iterable

        return it()


class CallableToIterable(Generic[_T]):
    def __init__(self, iterator_factory: Callable[[], Iterator[_T]]):
        self.iterator_factory = iterator_factory

    def __iter__(self) -> Iterator[_T]:
        return self.iterator_factory()



# tree.py
# credits to: https://stackoverflow.com/a/59109706/316766

from pathlib import Path

# prefix components:
space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


def tree(dir_path: Path, prefix: str = ''):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """
    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(path, prefix=prefix + extension)


def print_tree(path):
    for line in tree(Path(path)):
        print(line)


if __name__ == '__main__':
    print_tree(Path(__file__).parent.parent)



from __future__ import annotations

import os
from pathlib import Path

from wwwpy.server import configure


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--directory', '-d', default=os.getcwd(),
                        help='Specify alternative directory '
                             '[default:current directory]')
    parser.add_argument('port', action='store',
                        default=8000, type=int,
                        nargs='?',
                        help='Specify alternate port [default: 8000]')
    args = parser.parse_args()

    working_dir = Path(args.directory).absolute()
    configure.start_default(args.port, working_dir)


if __name__ == '__main__':
    main()



from inspect import iscoroutinefunction


async def entry_point():
    from wwwpy.common.tree import print_tree
    print_tree('/wwwpy_bundle')
    import js
    try:
        js.document.body.innerHTML = 'Going to import remote'
        import remote
        if hasattr(remote, 'main'):
            if iscoroutinefunction(remote.main):
                await remote.main()
            else:
                remote.main()
    except ImportError as e:
        import traceback
        msg = 'module remote load failed. Error: ' + str(e) + '\n\n' + traceback.format_exc() + '\n\n'
        js.document.body.innerHTML = msg.replace('\n', '<br>')



from __future__ import annotations

from pathlib import Path

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.common import iterlib
from wwwpy.resources import library_resources, from_filesystem, from_filepath
from wwwpy.rpc import Services, Module, Stubber
from wwwpy.webserver import Webserver, wait_forever
from wwwpy.webservers.available_webservers import available_webservers


def start_default(port: int, directory: Path):
    webserver = available_webservers().new_instance()

    convention(directory, webserver)

    webserver.set_port(port).start_listen()
    wait_forever()


def convention(directory, webserver):
    """
    Convention for a wwwpy server.
    It configures the webserver to serve the files from the working directory.
    It also configures the webserver to serve the files from the library.
    """
    print(f'applying convention to working_dir: {directory}')
    import sys
    sys.path.insert(0, str(directory))
    stubber_resources = []
    try:
        import server.rpc as rpc_module
        services = Services()
        rpc_module = Module(rpc_module)
        services.add_module(rpc_module)
        webserver.set_http_route(services.route)

        stubber_resources = [Stubber(services.route.path, services, rpc_module).remote_stub_resources()]
    except:
        pass

    resources = iterlib.repeatable_chain(
        library_resources(),
        from_filesystem(directory / 'remote', relative_to=directory),
        from_filesystem(directory / 'common', relative_to=directory),
        from_filepath(directory / 'common.py', relative_to=directory),
        from_filepath(directory / 'remote.py', relative_to=directory),
        *stubber_resources,
    )
    bootstrap_python = f'from wwwpy.remote.main import entry_point; await entry_point()'
    webserver.set_http_route(*bootstrap_routes(resources, python=bootstrap_python))

    return webserver



from __future__ import annotations

import inspect
import zipfile
from abc import ABC
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Iterator, Callable, Optional, TypeVar, Iterable
from zipfile import ZipFile

from wwwpy.common.iterlib import CallableToIterable

parent = Path(__file__).resolve().parent


@dataclass(frozen=True)
class Resource(ABC):
    """represents something (a file) to be transferred to the remote"""
    arcname: str


@dataclass(frozen=True)
class StringResource(Resource):
    content: str


@dataclass(frozen=True)
class PathResource(Resource):
    filepath: Path


# https://docs.python.org/3/library/collections.abc.html
# https://wiki.python.org/moin/Iterator
TResource = TypeVar("TResource", bound=Resource)

ResourceAccept = Callable[[TResource], bool]

_directory_blacklist = {'.mypy_cache', '__pycache__'}


def default_resource_accept(resource: Resource) -> bool:
    if not isinstance(resource, PathResource):
        return True
    filepath = resource.filepath
    if filepath.name == '.DS_Store':
        return False
    if filepath.is_dir() and filepath.name in _directory_blacklist:
        return False
    return True


def from_filepath(file: Path, relative_to: Path) -> Iterable[PathResource]:
    def bundle() -> Iterator[PathResource]:
        rel = file.relative_to(relative_to)
        if file.is_file() and file.exists():
            yield PathResource(str(rel), file)

    return CallableToIterable(bundle)


def _recurse(path: Path, relative_to: Path, resource_accept: ResourceAccept) -> Iterator[PathResource]:
    for f in path.glob('*'):
        rel = f.relative_to(relative_to)
        candidate = PathResource(str(rel), f)
        if resource_accept(candidate):
            if f.is_file():
                yield candidate
            else:
                yield from _recurse(f, relative_to, resource_accept)


def from_filesystem(
        folder: Path, relative_to: Path | None = None,
        resource_accept: ResourceAccept = default_resource_accept
) -> Iterable[PathResource]:
    relative_to_defined: Path = folder if relative_to is None else relative_to

    def bundle() -> Iterator[PathResource]:
        yield from _recurse(folder, relative_to_defined, resource_accept)

    return CallableToIterable(bundle)


def build_archive(resource_iterator: Iterator[Resource]) -> bytes:
    """builds a zip archive from the given resources and returns the bytes"""
    stream = BytesIO()
    zip_file = ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1)
    for resource in resource_iterator:
        if isinstance(resource, PathResource):
            zip_file.write(resource.filepath, resource.arcname)
        elif isinstance(resource, StringResource):
            zip_file.writestr(resource.arcname, resource.content)
        else:
            raise Exception(f'Unhandled class \n  type={type(resource).__name__} \n  data={resource}')
    zip_file.close()

    stream.seek(0)
    return stream.getbuffer().tobytes()


def stacktrace_pathfinder(stack_backtrack: int = 1) -> Optional[Path]:
    wwwpy_root = parent

    for stack in inspect.stack():
        source_path = Path(stack.filename).resolve()
        if not _is_path_contained(source_path, wwwpy_root):
            stack_backtrack -= 1
            if stack_backtrack == 0:
                return source_path

    return None


def _is_path_contained(child: Path, parent: Path) -> bool:
    cl = len(child.parts)
    cp = len(parent.parts)
    if cl < cp:
        return False
    m = min(cl, cp)
    child_parts = child.parts[:m]
    parent_parts = parent.parts[:m]
    return child_parts == parent_parts


def library_resources() -> Iterable[PathResource]:
    """returns the resources from the wwwpy library itself"""
    return from_filesystem(parent, relative_to=parent.parent)



import json
import traceback
from inspect import getmembers, isfunction, signature, iscoroutinefunction, Signature
from types import ModuleType, FunctionType
from typing import NamedTuple, List, Tuple, Any, Optional, Dict, Callable, Awaitable, Iterable, Iterator

from wwwpy.common.iterlib import CallableToIterable
from wwwpy.exceptions import RemoteException
from wwwpy.http import HttpRoute, HttpResponse, HttpRequest
from wwwpy.resources import Resource, StringResource

try:
    from typing_extensions import Protocol
except:
    try:
        from typing import Protocol
    except:
        class Protocol:
            ...

from wwwpy.unasync import unasync


class Function(NamedTuple):
    name: str
    func: FunctionType
    signature: str
    is_coroutine_function: bool
    sign: Signature
    blocking: FunctionType


def _std_function_to_function(fun_tuple: Tuple[str, FunctionType]) -> Function:
    name = fun_tuple[0]
    func = fun_tuple[1]
    sign = signature(func)
    is_coroutine_function = iscoroutinefunction(func)
    blocking = unasync(func)
    return Function(name, func, str(sign), is_coroutine_function, signature(func), blocking)


class Module:
    def __init__(self, module: ModuleType):
        self.module = module
        self.name = module.__name__

        self.functions: List[Function] = function_list(self.module)
        self._funcs = {f.name: f for f in self.functions}

    def __getitem__(self, name) -> Function:
        return self._funcs.get(name, None)


def function_list(module: ModuleType) -> List[Function]:
    return list(map(_std_function_to_function, getmembers(module, isfunction)))


class RpcResponse(NamedTuple):
    result: Any
    exception: str

    @classmethod
    def from_json(cls, string: str) -> 'RpcResponse':
        obj = json.loads(string)
        response = RpcResponse(*obj)
        return response

    def to_json(self) -> str:
        return json.dumps(self, default=str)


class RpcRequest(NamedTuple):
    module: str
    func: str
    args: List[Optional[Any]]

    def json(self) -> str:
        return json.dumps(self)

    @classmethod
    def build_request(cls, module_name: str, func_name: str, *args) -> 'RpcRequest':
        return RpcRequest(module_name, func_name, args)

    @classmethod
    def from_json(cls, string: str) -> 'RpcRequest':
        obj = json.loads(string)
        request = RpcRequest(*obj)
        return request


Fetch = Callable[[str, str, str], Awaitable[str]]


class Fetch(Protocol):
    def __call__(self, url: str, method: str = '', data: str = '') -> Awaitable[str]: ...


class Proxy:
    def __init__(self, module_name: str, rpc_url: str, fetch: Fetch):
        self.rpc_url = rpc_url
        self.fetch = fetch
        self.module_name = module_name

    async def dispatch(self, func_name: str, *args) -> Any:
        rpc_request = RpcRequest.build_request(self.module_name, func_name, *args)
        json_response = await self.fetch(self.rpc_url, method='POST', data=rpc_request.json())
        response = RpcResponse.from_json(json_response)
        ex = response.exception
        if ex is not None and ex != '':
            raise RemoteException(ex)
        return response.result


class Services:
    def __init__(self):
        self._modules: Dict[str, Module] = {}
        self.route = HttpRoute('/wwwpy/rpc', self._route_callback)

    def add_module(self, module: Module):
        self._modules[module.name] = module

    def find_module(self, module_name: str) -> Optional[Module]:
        return self._modules.get(module_name, None)

    def dispatch(self, request: str) -> str:
        rpc_request = RpcRequest.from_json(request)
        module = self.find_module(rpc_request.module)
        function = module[rpc_request.func]
        exception = ''
        result = None
        try:
            result = function.blocking(*rpc_request.args)
        except Exception:
            exception = traceback.format_exc()

        response = RpcResponse(result, exception)
        return response.to_json()

    def _route_callback(self, request: HttpRequest) -> HttpResponse:
        resp = self.dispatch(request.content)
        response = HttpResponse(resp, 'application/json')
        return response


class Stubber:
    def __init__(self, rpc_url: str, services: Services, rpc_module: Module = None):
        self._services = services
        self._rpc_url = rpc_url
        self._module: Optional[Module] = rpc_module

    def remote_stub_resources(self) -> Iterable[Resource]:
        def bundle() -> Iterator[Resource]:
            if self._module is None:
                return
            imports = 'from wwwpy.remote.fetch import async_fetch_str'
            stub_source = generate_stub_source(self._module, self._rpc_url, imports)
            yield StringResource(self._module.name.replace('.', '/') + '.py', stub_source)

        return CallableToIterable(bundle)

def generate_stub_source(module: Module, rpc_url: str, imports: str):
    module_name = module.name
    # language=python
    stub_header = f"""
from __future__ import annotations
from wwwpy.rpc import Proxy
{imports}

rpc_url = '{rpc_url}'
module_name = '{module_name}'
proxy = Proxy(module_name, rpc_url, async_fetch_str)
    """

    stub_functions = ''
    for f in module.functions:
        parameters = f.sign.parameters.values()
        params_list = ', '.join(p.name for p in parameters)
        args_list = '' if params_list == '' else ', ' + params_list
        fun_stub = f'\nasync def {f.name}{f.signature}:\n' + \
                   f'    return await proxy.dispatch("{f.name}"{args_list})\n'
        stub_functions += fun_stub

    return stub_header + stub_functions



from __future__ import annotations

from functools import partial
from http import HTTPStatus
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from threading import Thread
from typing import Optional, Callable, Dict, AnyStr, Tuple, Any
from urllib.parse import urlparse, parse_qs

from wwwpy.http import HttpRoute, HttpResponse, HttpRequest
from ..webserver import Webserver
from ..webserver import wait_forever


class WsPythonEmbedded(Webserver):
    def __init__(self) -> None:
        super().__init__()
        self.thread: Thread | None = None
        self._routes: Dict[str, HttpRoute] = {}

    def _setup_route(self, route: HttpRoute) -> None:
        self._routes[route.path] = route

    def _start_listen(self) -> None:
        httpd = ThreadingHTTPServer((self.host, self.port), partial(RequestHandler, handler=self._handler))

        def run() -> None:
            print(f'Starting embedded python web server on:\n'
                  f' - http://{self.host}:{self.port}\n'
                  f' - {self.localhost_url()}\n')
            httpd.serve_forever()

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()

    def _handler(self, request: 'RequestHandler') -> bool:
        params, rpath = request.decode_request()
        route = self._routes.get(rpath, None)
        if route is None:
            nf = HTTPStatus.NOT_FOUND
            request.send_bytes(bytes(nf.phrase, 'utf8'), code=nf.value)
        else:
            req = HttpRequest(request.command, request.get_body(), request.get_content_type())
            resp = route.callback(req)
            content = resp.content
            if isinstance(content, str):
                content = bytes(content, 'utf8')
            if not isinstance(content, bytes):
                raise Exception(f'type of the content not supported: {type(content)}')
            request.send_bytes(content, content_type=resp.content_type)
        return True


class RequestHandler(SimpleHTTPRequestHandler):
    query_index = 4
    path_index = 2

    def __init__(self, *args: Any
                 , handler: Callable[['RequestHandler'], bool]
                 , **kwargs: Any) -> None:
        self.handler = handler
        super(RequestHandler, self).__init__(*args, **kwargs)

    def do_GET(self) -> None:
        self.handler(self)

    def do_POST(self) -> None:
        self.handler(self)

    def get_content_type(self) -> str:
        return self.headers.get('Content-Type', '')

    def get_body(self) -> bytes:
        if self.command != 'POST':
            return b''
        content_len = int(self.headers.get('Content-Length', '0'))
        body = self.rfile.read(content_len)
        return body

    def decode_request(self) -> Tuple[Dict[str, str], str]:
        p = urlparse(self.path)
        query_str = p[self.query_index]
        rpath: str = p[self.path_index]
        di = parse_qs(query_str)
        params = {k: v[0] for k, v in di.items()}
        return params, rpath

    def send_bytes(self, content: bytes, code: int = 200, content_type: str = 'text/plain') -> int:
        self.protocol_version = "HTTP/1.1"
        self.send_response(code)
        self.send_header("Content-Type", content_type)
        self.send_header("Content-Length", str(len(content)))
        self.send_header("Access-Control-Allow-Headers", "*")
        self.send_header("Access-Control-Allow-Methods", "*")
        self.send_header("Access-Control-Allow-Origin", "*")
        self.end_headers()
        return self.wfile.write(content)

    def serve_file(self, directory: str, filename: str) -> None:
        self.directory = directory
        self.path = '/' + filename
        super(RequestHandler, self).do_GET()


if __name__ == '__main__':
    s = WsPythonEmbedded()
    s.set_http_route(HttpRoute('/', lambda req: HttpResponse('ciao', 'text/html')))
    s.start_listen()
    wait_forever()



from __future__ import annotations

import datetime as datetime_mod
import time
from datetime import datetime, date, timedelta
from typing import Union, Optional, Any, TypeVar, Callable

from mypy_extensions import VarArg, KwArg

# FuncT = TypeVar("FuncT", bound=Callable[[VarArg(Any), KwArg(Any)], Any])
FuncT = Callable[[VarArg(Any), KwArg(Any)], Any]


def time_wrapper(above_seconds: float = 0.0, callback: FuncT | None = None, *outer_args: Any,
                 **outer_kwargs: Any) -> FuncT:
    def decorator(function: FuncT) -> FuncT:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            start = time.time()
            result = function(*args, **kwargs)
            end = time.time()
            delta_secs = (end - start)
            if delta_secs >= above_seconds:
                if callback is None:
                    print('{:s} function took {:.3f} ms'.format(function.__name__, delta_secs * 1000.0))
                else:
                    callback(delta_secs, *args)
            return result

        return wrapper

    return decorator


# def time_wrapper_old(f):
#     def wrapper(*args, **kwargs):
#         start = time.time()
#         ret = f(*args, **kwargs)
#         end = time.time()
#         print('{:s} function took {:.3f} ms'.format(f.__name__, (end - start) * 1000.0))
#         return ret
#
#     return wrapper


def wait_until(end: Union[date, datetime]) -> None:
    """
    When the function ends, it will be a little after `end`
    """
    if isinstance(end, date) and not isinstance(end, datetime):
        end = datetime.combine(end, datetime.min.time())
    if not isinstance(end, datetime):
        raise Exception(f'Argument needs to be a date or a datetime. It is a type={type(end)} value={end}')

    while True:
        diff = (end - datetime.now()).total_seconds()
        if diff < 0:  # will return after now is after the wait point in time
            return
        print(f'wait_until({end}) diff={diff} sleeping diff/2...')
        time.sleep(diff / 2)


def first_day_next_month(today: Optional[date] = None) -> date:
    if today is None:
        today = date.today()
    dt = today
    while dt.day != 1 or dt.month == today.month:
        dt = dt + timedelta(days=1)
    return dt


def main_harness() -> None:
    now = datetime.now()
    nt = now + timedelta(seconds=2)
    new_time = datetime_mod.time(nt.hour, nt.minute, nt.second)
    dt = datetime.combine(now.date(), new_time)
    wait_until(dt)
    now = datetime.now()
    print(f'one second is passed diff={dt - now}  dt={dt} now={now}')
    dt = datetime.combine(first_day_next_month(), datetime_mod.time(9, 30, 0))
    print(dt)
    wait_until(dt)


if __name__ == '__main__':
    main_harness()



from __future__ import annotations
from typing import Iterator, List

from ..webserver import Webserver


class AvailableWebservers:
    def __init__(self) -> None:
        self.classes = _webservers_classes()
        self.ids = list(map(lambda w: w.__name__, _webservers_classes()))

    def new_instance(self) -> Webserver:
        return self.classes[0]()

    def instances(self) -> Iterator[Webserver]:
        for webserver_class in self.classes:
            yield webserver_class()


def _webservers_classes() -> List[type[Webserver]]:
    result: List[type[Webserver]] = []
    try:
        from .fastapi import WsFastapi
        result.append(WsFastapi)
    except:
        pass
    try:
        from .flask import WsFlask
        result.append(WsFlask)
    except:
        pass
    try:
        from .tornado import WsTornado
        result.append(WsTornado)
    except:
        pass

    from .python_embedded import WsPythonEmbedded
    result.append(WsPythonEmbedded)

    return result


_available_webservers: AvailableWebservers | None = None


def available_webservers() -> AvailableWebservers:
    global _available_webservers
    if _available_webservers is None:
        _available_webservers = AvailableWebservers()
    return _available_webservers
